"""Terminal interface and other UI functions."""

import sys
import time

from cvsgit.i18n import _

class Progress(object):
    """Display progress information.
    """

    def __init__(self):
        self.last_progress = 0
        self.last_message = ''
        self.last_count = None
        self.last_total = None
        self.update_suppressed = False

        if sys.stdout.isatty():
            self.update_interval = 1
            self.update = self.update_tty
            self.finish = self.finish_tty
        else:
            self.update_interval = 30
            self.update = self.update_dumb
            self.finish = self.finish_dumb

    def __enter__(self):
        self.last_progress = time.time()

    def __exit__(self, exception_type, value, traceback):
        self.last_progress = 0
        if self.update_suppressed:
            self.update(self.last_message, self.last_count, self.last_total)
            self.update_suppressed = False
        self.finish()
        self.last_message = None
        return False

    def __call__(self, message, count=None, total=None):
        if (self.last_message and message != self.last_message) or \
                time.time() - self.last_progress > self.update_interval:
            self.last_progress = time.time()
            self.update(message, count, total)
            self.update_suppressed = False
        else:
            self.update_suppressed = True
        self.last_message = message
        self.last_count = count
        self.last_total = total

    def update_tty(self, message, count, total):
        if self.last_message:
            sys.stdout.write('\r' + (' ' * len(self.last_message)) + '\r')
        if count == None:
            sys.stdout.write('%s...' % message)
        elif total == None:
            sys.stdout.write('%s: %d' % (message, count))
        elif count == total:
            sys.stdout.write('%s: %s (%d/%d)' % \
                (message, _('done.'), count, total))
        else:
            sys.stdout.write('%s: %3.0f%% (%d/%d)' % \
                (message, count * 100.0 / total, count, total))
        sys.stdout.flush()

    def finish_tty(self):
        sys.stdout.write('\n')

    def update_dumb(self, message, count, total):
        if count == None:
            sys.stdout.write('%s...\n' % message)
        elif total == None:
            sys.stdout.write('%s: %d\n' % (message, count))
        elif count == total:
            sys.stdout.write('%s: %s (%d/%d)\n' % \
                (message, _('done.'), count, total))
        else:
            sys.stdout.write('%s: %3.0f%% (%d/%d)\n' % \
                (message, count * 100.0 / total, count, total))
        sys.stdout.flush()

    def finish_dumb(self):
        pass

class NoProgress(object):
    """Behaves like Progress but does nothing.
    """

    def __enter__(self):
        pass

    def __exit__(self, exception_type, value, traceback):
        return False

    def __call__(self, message, count=None, total=None):
        pass



"""RCS interface module for CVSGit."""

from __future__ import absolute_import
import rcsparse

# Some observations about RCS + CVS, although I don't really know
# that much about the RCS format or the CVS usage of it...
#
# 1. The 'branch' keyword is normally absent (or empty?), but has
#    the value "1.1.1" for files that have been imported into the
#    vendor branch and never modified.  The 'head' keyword has the
#    value "1.1" in that case.
#
# 2. When checking out files on trunk via date that were imported
#    in a vendor branch, cvs expands the Id keyword to "1.1.1.1"
#    if the date matches exactly, and to "1.1" if the date is one
#    second or more after the exact date of the import.

import os.path
import sys

from cvsgit.changeset import Change, FILE_ADDED, FILE_MODIFIED, \
    FILE_DELETED

from cvsgit.error import Error
from cvsgit.i18n import _

REV_TIMESTAMP = 1
REV_AUTHOR = 2
REV_STATE = 3
REV_BRANCHES = 4
REV_NEXT = 5
REV_MODE = 6

class RCSError(Error):
    """Base class for exceptions from the cvsgit.rcs module.
    """

class ParseError(RCSError):
    """Raised when an RCS file couldn't be parsed correctly.
    """

    def __init__(self, message, rcsfile):
        """This exception provides additional information.

        'rcsfile' is an RCSFile object.
        """
        super(ParseError, self).__init__(message)
        self.rcsfile = rcsfile

class CheckoutError(ParseError):
    """Extraction of fulltext from RCS file failed.

    This exception indicates that the fulltext of a particular
    revision was requested but couldn't be retrieved from the
    RCS file because the revision does not exist, for example.

    It likely indicates that an invalid argument was passed to
    the RCSFile.blob method.
    """

    def __init__(self, rcsfile, revision):
        """This exception provides additional information.

        'rcsfile' is an RCSFile object.
        'revision' is the revision that couldn't be retrieved.
        """
        super(CheckoutError, self).__init__(
            _("Couldn't retrieve file content for revision %s of %s") % \
                (revision, rcsfile.filename), rcsfile)
        self.revision = revision

class RCSFile(object):
    """Represents a single RCS file.
    """

    def __init__(self, filename, encoding='iso8859-1'):
        """'encoding' sets the encoding assumed of log messages and
        delta text in RCS files.
        """
        self.filename = filename
        self.encoding = encoding
        self.rcsfile = rcsparse.rcsfile(filename)

    head = property(lambda self: self.rcsfile.head)
    branch = property(lambda self: self.rcsfile.branch)
    expand = property(lambda self: self.rcsfile.expand)
    mode = property(lambda self: self.rcsfile.mode)
    revs = property(lambda self: self.rcsfile.revs)

    def revisions(self):
        """Yield all revision numbers from current HEAD backwards.
        """
        if self.branch:
            branchprefix = self.branch + '.'
        else:
            branchprefix = None

        revision = self.head
        while revision != None:
            if branchprefix:
                for brevision in self.revs[revision][REV_BRANCHES]:
                    if brevision.startswith(branchprefix):
                        branchprefix = None
                        revision = brevision
                        break

            yield(revision)
            revision = self.revs[revision][REV_NEXT]

    def changes(self):
        """Yield Change objects for all revisions on HEAD

        The changes are generated by following the current head
        revision back to its origin.  The order of changes is thus
        from most recent to oldest.
        """
        for revision in self.revisions():
            change = self.change(revision)
            if change != None:
                yield(change)

    def change(self, revision):
        """Return a single Change object for <revision>.

        If the revision is 1.1 and has state 'dead' then the file was
        added on a branch and None is returned.
        """
        rev = self.revs[revision]
        if rev[REV_STATE] == 'dead':
            if revision == '1.1':
                # This file was initially added on a branch and so
                # the initial trunk revision was marked 'dead'. We
                # do not count this as a change since it wasn't
                # added and hasn't existed before.
                return None
            else:
                filestatus = FILE_DELETED
        elif rev[REV_NEXT] == None:
            filestatus = FILE_ADDED
        else:
            # XXX: Resurrections of dead revisions aren't flagged
            # as FILE_ADDED.
            filestatus = FILE_MODIFIED

        # The log message for an initial import is actually in
        # the initial vendor branch revision.
        if revision == '1.1' and '1.1.1.1' in rev[REV_BRANCHES]:
            log = self.rcsfile.getlog('1.1.1.1')
        else:
            log = self.rcsfile.getlog(revision)

        # XXX: is this right?
        log = unicode(log, self.encoding)

        if rev[REV_MODE] == None:
            mode = ''
        else:
            mode = rev[REV_MODE]

        return Change(timestamp=rev[REV_TIMESTAMP],
                      author=rev[REV_AUTHOR],
                      log=log,
                      filestatus=filestatus,
                      filename=self.filename,
                      revision=revision,
                      state=rev[REV_STATE],
                      mode=mode)

    def blob(self, revision):
        """Returns the revision's file content.
        """
        try:
            return self.rcsfile.checkout(revision)
        except RuntimeError:
            raise CheckoutError(self, revision)

    # XXX only for debugging; remove later
    def _print_revision(self, revision):
        import time
        rev = self.revs[revision]
        print 'revision:', revision
        print '  timestamp:', time.strftime("%Y-%m-%d %H:%M", time.gmtime(rev[REV_TIMESTAMP]))
        print '  branches:', rev[REV_BRANCHES]
        print '  next:', rev[REV_NEXT]
        print '  state:', rev[REV_STATE]
        print '  log:', self.rcsfile.getlog(revision).splitlines()[0]



"""Command to dump changes fetched from CVS."""

import re
import subprocess
from subprocess import PIPE
import sys

from cvsgit.cvs import split_cvs_source
from cvsgit.i18n import _
from cvsgit.main import Command, Conduit
from cvsgit.utils import Tempdir, stripnl
from cvsgit.term import NoProgress, Progress

class DumpChanges(Command):
    __doc__ = _(
    """Dump changes fetched from CVS

    Usage: %prog

    Dumps all previously fetched changes from the CVS repository.
    """)

    def initialize_options(self):
        pass

    def finalize_options(self):
        if len(self.args) > 0:
            self.usage_error(_('too many arguments'))

    def run(self):
        cvs = Conduit().cvs
        for change in cvs.changes():
            print change



"""Main application class and user interface helper routines for
'git-cvs'."""

import os.path
import re

from cvsgit.cmd import Cmd
from cvsgit.error import Error
from cvsgit.git import Git
from cvsgit.cvs import CVS
from cvsgit.meta import MetaDb
from cvsgit.i18n import _
from cvsgit.term import Progress

class Command(Cmd):
    """Base class for conduit commands
    """

    def add_authors_option(self):
        self.add_option('--authors', metavar='AUTHORS', help=\
            _("Map CVS committer login names to fullnames."))

    def finalize_authors_option(self):
        if not self.options.authors:
            return
        authors = {}
        with open(self.options.authors, 'r') as file:
            while True:
                line = file.readline().strip()
                if not line:
                    break

                match = re.match('^([^\s=]+)(?:\s*=\s*|\s+)(.+?)(?:\s+<(.+)>)?$', line)
                if match:
                    login, fullname, email = match.groups()
                    authors[login] = (fullname, email)
                else:
                    self.warn('invalid line in authors map: %s' % line)
            self.options.authors = authors

    def add_stop_on_unknown_author_option(self):
        self.add_option('--stop-on-unknown-author', action='store_true', help=\
            _("Abort the operation if any author mapping is missing."))

    def add_quiet_option(self):
        self.add_option('--quiet', action='store_true', help=\
            _("Only report error and warning messages."))

    def add_verbose_option(self):
        self.add_option('--verbose', action='store_true', help=\
            _("Display each changeset as it is imported."))

    def add_no_skip_latest_option(self):
        self.add_option('--no-skip-latest', action='store_true', help=\
            _("Import potentially incomplete changesets instead of retaining them for the next incremental import."))

class ConduitError(Error):
    """Base exception for errors in the cvsgit.main module
    """

class NoSourceError(ConduitError):
    """Raised when no CVS source has been configured

    This error may occur if you run git-cvs commands in a directory
    that has not been initialized with "git-cvs init".
    """
    def __init__(self):
        super(NoSourceError, self).__init__(
            _("'cvs.source' is unset; not a git-cvs repository?"))

class UnknownAuthorFullnames(ConduitError):
    """Raised when there is no known fullname for an author's login
    name and --stop-on-missing-author was given on the command-line.
    """

    def __init__(self, authors):
        msg = 'unknown authors: %s' % ', '.join(authors)
        super(ConduitError, self).__init__(msg)

class Conduit(object):
    """CVS-to-Git conduit logic
    """

    def __init__(self, directory=None):
        self.git = Git(directory)
        self.branch = 'refs/heads/cvs/HEAD'
        self._cvs = None
        self._config = {}

    def config_get(self, varname):
        """Get a Git variable from the 'cvs' section
        """
        if self._config.has_key(varname):
            return self._config[varname]
        else:
            value = self.git.config_get('cvs.' + varname)
            self._config[varname] = value
            return value

    def config_set(self, varname, value):
        """Set a Git variable in the 'cvs' section
        """
        self.git.config_set('cvs.' + varname, value)
        self._config[varname] = value

    def get_source(self):
        """Get the CVS repository source path
        """
        source = self.config_get('source')
        if source is None:
            raise NoSourceError
        return source

    def set_source(self, directory):
        """Set the CVS repository source path
        """
        self.config_set('source', directory)

    source = property(get_source, set_source)

    def get_domain(self):
        return self.config_get('domain')

    def set_domain(self, directory):
        self.config_set('domain', directory)

    domain = property(get_domain, set_domain)

    def get_cvs(self):
        if self._cvs == None:
            filename = os.path.join(self.git.git_dir, 'cvsgit.db')
            metadb = MetaDb(filename)
            self._cvs = CVS(self.source, metadb)
        return self._cvs

    cvs = property(get_cvs)

    def init(self, repository, domain=None, bare=False, quiet=True):
        self.git.init(bare=bare, quiet=quiet)

        if not self.git.is_bare() and \
                self.git.config_get('branch.master.remote') == None:
            self.git.config_set('branch.master.remote', '.')
            self.git.config_set('branch.master.merge', self.branch)
            self.git.config_set('branch.master.rebase', 'true')

        self.source = repository

        if domain:
            self.domain = domain

    def fetch(self, limit=None, quiet=True, verbose=False,
              flush=False, authors=None, stop_on_unknown_author=False):
        """Fetch new changesets into the CVS tracking branch.
        """
        if quiet or verbose:
            progress = None
        else:
            progress = Progress()

        self.cvs.fetch(progress=progress, limit=limit, flush=flush)

        # XXX: Should not access private self.cvs.metadb.
        if authors and stop_on_unknown_author:
            unknown = []
            for author in self.cvs.metadb.all_authors():
                if not authors.has_key(author):
                    unknown.append(author)
            if len(unknown) > 0:
                raise UnknownAuthorFullnames(unknown)

        self.git.import_changesets(self.cvs.changesets(), self.branch,
                                   domain=self.domain,
                                   limit=limit,
                                   verbose=verbose,
                                   progress=progress,
                                   total=self.cvs.count_changesets(),
                                   authors=authors,
                                   stop_on_unknown_author=\
                                       stop_on_unknown_author)

    def pull(self, limit=None, quiet=True, verbose=False, flush=False,
             authors=None, stop_on_unknown_author=False):
        self.fetch(limit=limit, quiet=quiet, verbose=verbose,
                   flush=flush, authors=authors, stop_on_unknown_author=
                   stop_on_unknown_author)

        args = []
        if quiet:
            args.append('--quiet')

        if self.git.is_bare():
          self.git.check_command('branch', '-f', 'master', self.branch)
        else:
          # XXX: --quiet is not enough if branch.<branch>.rebase is true
          #self.git.pull(*args)
          import subprocess
          self.git.check_command('pull', *args, stdout=subprocess.PIPE)



"""CVS interface for CVSGit."""

import os.path
import re
import time

from subprocess import Popen, PIPE

from cvsgit.changeset import ChangeSetGenerator
from cvsgit.rcs import RCSFile
from cvsgit.i18n import _
from cvsgit.term import NoProgress
from cvsgit.utils import stripnl

def split_cvs_source(dirname):
    """Split <dirname> into CVSROOT and module paths.
    """
    cvsroot = dirname
    module = ''
    while True:
        parent = os.path.dirname(cvsroot)
        if cvsroot == parent:
            raise TypeError, _('not a CVS repository path (%s): %s') \
                % (_('no CVSROOT within nor above'), dirname)
        if os.path.isdir(os.path.join(cvsroot, 'CVSROOT')):
            return (cvsroot, module,)
        if module == '':
            module = os.path.basename(cvsroot)
        else:
            module = os.path.join(os.path.basename(cvsroot), module)
        cvsroot = parent

class CVS(object):
    """Represents a CVS repository.
    """

    def __init__(self, dirname, metadb):
        self.metadb = metadb

        # 'dirname' is a local filesystem path pointing at the root of
        # a CVS repository or at a module within.  If it is a module
        # path, operations will be limited to that module.  Otherwise,
        # the module path will be empty and operations will by default
        # apply to the whole repository.
        #
        # In any case, the repository path is always determined by
        # recursively following the parents of 'dirname' until a a
        # sub-directory named 'CVSROOT' is found.  Repository path and
        # module path are available for reading in the instance
        # attributes 'root' and 'module', respectively.

        if not os.path.isdir(dirname):
            raise TypeError, _('not a CVS repository path (%s): %s') \
                % (_('not even a directory'), dirname)

        # Convert to absolute pathname, so that our logic doesn't
        # break if anyone uses os.chdir() and so that the path
        # traversal below can always assume an absolute path.
        dirname = os.path.abspath(dirname)

        # Split 'dirname' into self.root and self.module and also
        # set self.prefix to the full absolute module path.
        self.root, self.module = split_cvs_source(dirname)
        if self.module == '':
            self.prefix = self.root
        else:
            self.prefix = os.path.join(self.root, self.module)

        self.localid = None
        self.parse_config()

        self.statcache = {}
        self._rcs_log_keyword_re = re.compile('(.*)\$Log(?::[^$\r\n]+)?\$(.*)')
        self._rcs_keyword_re = re.compile('\$([A-Z][A-Za-z]+)(:[^$\r\n]*)?\$')
        self._rcs_strip_attic_re = re.compile('(Attic/)?([^/]+),v$')

    def parse_config(self):
        """Extract relevant information from the CVSROOT/config file,
        if it exsits.  At the moment, the only information that is
        used from the file is the custom Id tag keyword 'tag'."""

        filename = os.path.join(self.root, 'CVSROOT', 'config')
        if not os.path.isfile(filename):
            return

        f = file(filename, 'r')
        try:
            for line in f.readlines():
                # Skip empty and comment-only lines.
                if re.match('^\s*(#.*)?$', line) != None:
                    continue

                # Remaining lines must be "key=value" pairs.  split() will
                # fail otherwise.
                option, value = line.split('=', 2)
                if option == 'tag':
                    self.localid = value.strip()
        finally:
            f.close()

    # Helper function to check with the statcache and the stat()
    # system call if a file or directory is unmodified.
    def _unmodified(self, path):
        st = os.stat(os.path.join(self.prefix, path))
        identity = (st.st_mtime, st.st_size,)
        return self.statcache.has_key(path) and \
                self.statcache[path] == identity

    def changed_rcs_filenames(self, progress=None):
        """Return the list of RCS filenames which need to be scanned for
        new changes to import.
        """
        if not progress:
            progress = NoProgress()

        with progress:
            return self._changed_rcs_filenames(progress=progress)

    def _changed_rcs_filenames(self, progress=None):
        # Helper function to raise the OSError reported by os.walk().
        def raise_error(e): raise e

        self.statcache = self.metadb.load_statcache()
        result = []
        count = 0

        for dirpath, dirnames, filenames in \
                os.walk(self.prefix, onerror=raise_error):

            # Convert from absolute to relative path.
            assert(dirpath.startswith(self.prefix))
            dirpath = dirpath[len(self.prefix)+1:]

            # TODO: fill stat() cache for directories
            #if self._unmodified(dirpath):
            #    continue

            # Are we in an Attic directory? Then we must check each
            # filename for a "zombie" copy in the parent directory
            # and decide which one to use.
            in_attic = os.path.basename(dirpath) == 'Attic'
            if in_attic:
                parent = os.path.dirname(dirpath)

            for filename in filenames:
                # Ignore all non-RCS files.
                if not filename.endswith(',v'):
                    continue

                count += 1
                progress(_('Collecting RCS files'), count)

                #
                # Perform the zombie check:
                #
                # 1.) If the zombie is in the parent directory, remove
                #     it from 'result' and add the one in the Attic.
                #
                # 2.) If the zombie is in the Attic, leave the good one
                #     in 'result' and ignore the one in the Attic.
                #
                # 3.) If neither of the two files can be classified as
                #     a zombie, raise an error.
                #
                if in_attic and self._zombie_check(result, parent, filename):
                    # This is case 2.) above: skip the Attic filename.
                    continue

                filename = os.path.join(dirpath, filename)
                if not self._unmodified(filename):
                    result.append(filename)

        return result

    def _zombie_check(self, result, parent, filename):
        """Check a path for zombie files.  If a path exists in the Attic
        and the parent directory, one of them must be a zombie copy.  If
        it cannot be determined which one is the zombie and which one is
        the real copy, raise an error.

        This function should only be called during an os.walk() run when
        searching for files an Attic directory.  This guarantees that we
        have already seen the other copy in the parent directory, if one
        exists.  The return value is True if the zombie is in the Attic
        and False if the zombie is in the parent directory."""

        trunkfile = os.path.join(parent, filename)
        if not os.path.isfile(trunkfile):
            # No zombie present; the file exists only in Attic.
            return False

        atticfile = os.path.join(parent, 'Attic', filename)
        # FIXME: Not a reliable test. We should make sure that the
        # zombie contains a subset of the revisions of the real copy.
        if os.path.getsize(trunkfile) < os.path.getsize(atticfile):
            result.remove(trunkfile)
            return False

        raise RuntimeError, \
            _("invalid path: %s (%s)") % (trunkfile, \
            _('exists in Attic and parent directory'))

    def fetch_changes(self, progress=None):
        """Fetch new revisions from the CVS repository.
        """
        if progress == None:
            progress = NoProgress()

        filenames = self.changed_rcs_filenames(progress=progress)
        with progress:
            self._fetch_changes(filenames, progress)

    def _fetch_changes(self, filenames, progress):
        count = 0
        total = len(filenames)
        progress(_('Parsing RCS files'), count, total)

        # We will commit changes to the database every few seconds to
        # avoid having to scan all RCS files again in case the process
        # is interrupted or an error occurs (a rescan isn't really bad
        # but costs time.)
        commit_time = 0
        commit_interval = 10
        self.metadb.begin_transaction()
        for rcsfile in filenames:
            try:
                self._fetch_changes_from_rcsfile(rcsfile)

                if time.time() - commit_time >= commit_interval:
                    try:
                        self.metadb.end_transaction()
                    except:
                        pass
                    self.metadb.begin_transaction()
                    commit_time = time.time()

                count += 1
                progress(_('Parsing RCS files'), count, total)
            except KeyboardInterrupt:
                # Re-raise the exception silently.  An impatient user
                # may interrupt the process and that should by handled
                # gracefully.
                raise
            except:
                # Print the file name where this error happened,
                # regardless of whether the error is actually printed,
                # just as a quick & dirty debugging aid.
                # TODO: raise a FetchChangesError
                print "(Error while processing %s)" % rcsfile
                raise
            finally:
                try:
                    self.metadb.end_transaction()
                except:
                    pass

    def _fetch_changes_from_rcsfile(self, rcsfile):
        # For the working copy path it does not matter if the RCS
        # file is in the 'Attic' directory or not, so strip it.
        filename = self._rcs_strip_attic_re.sub('\\2', rcsfile)

        abspath = os.path.join(self.prefix, rcsfile)
        st = os.stat(abspath)
        identity = (st.st_mtime, st.st_size,)

        for change in RCSFile(abspath).changes():
            # Record the file's actual working copy path, which
            # RCS alone cannot know about.
            change.filename = filename
            self.metadb.add_change(change)

        self.metadb.update_statcache({rcsfile:identity})

    def generate_changesets(self, progress=None, limit=None, flush=False):
        """Convert changes stored in the meta database into sets of
        related changes and store the resulting changesets in the meta
        database as well.

        If limit is given, only that many changesets will be imported and
        all others will be retained.  This can be used to import changesets
        in small batches.

        If flush is True, all changesets produced by the ChangeSetGenerator
        are imported; otherwise, potentially incomplete changesets may be
        retained for the next incremental import.  Use this flag if you can
        be sure that the CVS repository is consistent and is not going to be
        modified during the import.
        """
        if progress == None:
            progress = NoProgress()

        with progress:
            count = 0
            csg = ChangeSetGenerator(limit=limit)
            total = self.metadb.count_changes()
            progress(_('Processing changes'), 0, total)

            for change in self.changes(processed=False, reentrant=True):
                count += 1
                progress(_('Processing changes'), count, total)
                for cs in csg.integrate(change):
                    self.metadb.add_changeset(cs)

        if flush:
            # All changesets are assumed to be complete and will be
            # imported.  Note that the ChangeSetGenerator still counts
            # changesets from flush() against the specified limit.
            for cs in csg.flush():
                self.metadb.add_changeset(cs)
        elif len(csg.changesets) > 0:
            # The ChangeSetGenerator retained some changesets because
            # they are potentially incomplete.
            progress(_('Retained changesets'), len(csg.changesets))

    def fetch(self, progress=None, limit=None, flush=False):
        """Fetch new revisions and compute changesets.
        """
        self.fetch_changes(progress)
        self.generate_changesets(progress, limit, flush)

    def changesets(self):
        """Yield new changesets computed earlier.
        """
        for changeset in self.metadb.changesets_by_start_time():
            changeset.provider = self
            yield(changeset)

    def changes(self, processed=None, reentrant=True):
        """Yields changes fetched earlier.

        The 'processed' keyword can be set to a boolean value to
        select whether changes which are already included in a
        previously computed changeset are to be considered.  If the
        value is None, then all changes are considered.
        """
        return self.metadb.changes_by_timestamp(processed=processed,
                                                reentrant=reentrant)

    def rcsfilename(self, change):
        """Return the RCS filename corresponding to <change>.
        """
        filename = change.filename + ',v'

        result = os.path.join(self.prefix, filename)
        if os.path.isfile(result):
            return result

        result = os.path.join(self.prefix,
                              os.path.dirname(filename), 'Attic',
                              os.path.basename(filename))
        if os.path.isfile(result):
            return result

        raise RuntimeError, _('no RCS file found for %s') % \
            change.filename

    def rcsfile(self, change):
        """Return an RCSFile object for <change>.
        """
        return RCSFile(self.rcsfilename(change))

    def perm(self, change):
        """Return the file permissions as a decimal number.
        """
        return os.stat(self.rcsfilename(change)).st_mode & 0777

    def blob(self, change, changeset):
        """Return the raw binary content of a file at the specified
        revision.
        """
        filename = change.filename + ',v'
        rcsfile = os.path.join(self.prefix, filename)
        if not os.path.isfile(rcsfile):
            rcsfile = os.path.join(self.prefix,
                os.path.dirname(filename), 'Attic',
                os.path.basename(filename))
        if not os.path.isfile(rcsfile):
            raise RuntimeError, _('no RCS file found for %s') % filename

        # cvs has the odd behavior that it favors revision 1.1 over
        # 1.1.1.1 if it searches for a revision by date and the date
        # is after that of the initial revision even by one second.
        #if change.revision == '1.1.1.1' and \
        #        change.timestamp < changeset.timestamp:
        #    revision = '1.1'
        #else:
        #    revision = change.revision
        revision = change.revision

        rcsfile = RCSFile(rcsfile)
        blob = rcsfile.blob(revision)
        return self.expand_keywords(blob, change, rcsfile, revision)

    def expand_keyword(self, change, rcsfile, revision, kw):
        if kw == 'Id' or (self.localid and kw == self.localid):
            return ('%s %s %s %s %s' %
                    (os.path.basename(rcsfile.filename),
                     revision, time.strftime('%Y/%m/%d %H:%M:%S',
                     time.gmtime(change.timestamp)), change.author,
                     change.state))
        elif kw == 'Header':
            return ('%s %s %s %s %s' %
                    (rcsfile.filename, revision,
                     time.strftime('%Y/%m/%d %H:%M:%S',
                     time.gmtime(change.timestamp)), change.author,
                     change.state))
        elif kw == 'Date':
            return time.strftime('%Y/%m/%d %H:%M:%S',
                                 time.gmtime(change.timestamp))
        elif kw == 'Revision':
            return revision
        elif kw == 'Source':
            return rcsfile.filename
        elif kw == 'Author':
            return change.author
        elif kw in 'RCSfile':
            return os.path.basename(rcsfile.filename)
        elif kw == 'Log':
            # additional lines get inserted elsewhere
            return os.path.basename(rcsfile.filename)
        elif kw == 'State':
            return change.state
        elif kw in ('Locker', 'Name'):
            return ''
        elif kw == 'Mdocdate': # for OpenBSD
            timestamp = time.gmtime(change.timestamp)
            mdocdate = time.strftime('%B %e %Y', timestamp)
            mdocdate = mdocdate.replace('  ', ' ') # for %e
            return mdocdate

    def expand_log_keyword(self, change, rcsfile, revision, prefix):
        timestamp = time.gmtime(change.timestamp)
        timestamp = time.strftime('%Y/%m/%d %H:%M:%S', timestamp)
        s = prefix + ('Revision %s  %s  %s\n' %
                      (revision, timestamp, change.author))
        log = stripnl(rcsfile.rcsfile.getlog(revision))
        for line in log.split('\n'):
            s += (prefix + line).rstrip() + '\n'
        s += prefix.rstrip()
        return s

    # Expand RCS keywords while avoiding to copy the whole file
    # content into a new object when there's nothing to expand.
    def expand_keywords(self, blob, change, rcsfile, revision):
        # Skip keyword expansion for binary mode files.
        if change.mode == 'b' or rcsfile.expand not in (None, '', 'kv'):
            return blob

        text = None
        start = blob.find('$')

        while start != -1:
            end = blob.find('$', start + 1)
            if end == -1:
                break

            s = blob[start:end + 1]
            m = self._rcs_keyword_re.match(s)
            if m:
                kw = m.group(1)
                s2 = self.expand_keyword(change, rcsfile, revision, kw)
                if s2 != None:
                    #print '%s: >>>%s<<<' % (rcsfile.filename, s2)
                    if text == None:
                        text = blob[0:start]
                    text += '$' + kw + ': ' + s2.encode('ascii') + ' '
                    if kw == 'Log':
                        end += 1
                        text += '$\n'
                        prefix = blob[blob.rfind('\n', 0, start) + 1:start]
                        text += self.expand_log_keyword(
                            change, rcsfile, revision, prefix
                            ).encode('ascii')
                elif text != None:
                    text += blob[start:end]
            elif text != None:
                text += blob[start:end]

            start = end

        if text != None:
            text += blob[start:]

        # Return the original input string or expanded version.
        if text == None:
            return blob
        else:
            return text

    def mark_changeset(self, changeset):
        """Mark 'changeset' as having been committed to "the other
        VCS"."""

        assert(changeset.mark != None)
        self.metadb.mark_changeset(changeset.id, changeset.mark)

    def count_changesets(self):
        return self.metadb.count_changesets()



"""Command to clone a CVS repository or module as a Git repository."""

import os.path
import shutil

from cvsgit.main import Command, Conduit
from cvsgit.i18n import _
from cvsgit.command.verify import Verify

class Clone(Command):
    __doc__ = _(
    """Clone a CVS repository or module into a Git repository.

    Usage: %prog [options] <repository> [<directory>]

    Clones an entire CVS repository or a module into a Git repository.
    The source argument <repository> must be a local path pointing at
    the CVS repository root or a module directory within.  The
    destination argument <directory> is selected automatically, based
    on the last component of the source path.
    """)

    def initialize_options(self):
        self.repository = None
        self.directory = None
        self.add_option('--bare', action='store_true', help=\
            _("Create a bare Git repository without work tree."))
        self.add_option('--limit', type='int', metavar='COUNT', help=\
            _("Stop importing after COUNT new commits."))
        self.add_option('--domain', metavar='DOMAIN', help=\
            _("Set the e-mail domain to use for unknown authors."))
        self.add_option('--verify', action='store_true', help=\
            _("Run the verify command after cloning (does not work "
              "with --bare)."))
        self.add_option('--no-repack', action='store_true', help=\
            _("Don't run \"git repack -adF\" after cloning (so you "
              "end up with an uncompressed pack file)."))
        self.add_quiet_option()
        self.add_verbose_option()
        self.add_no_skip_latest_option()
        self.add_authors_option()
        self.add_stop_on_unknown_author_option()

    def finalize_options(self):
        if len(self.args) < 1:
            self.usage_error(_('missing CVS repository path'))
        elif len(self.args) == 1:
            self.repository = os.path.abspath(self.args[0])
            self.directory = os.path.basename(self.repository)
        elif len(self.args) == 2:
            self.repository, self.directory = self.args
        else:
            self.usage_error(_('too many arguments'))

        self.finalize_authors_option()

    def run(self):
        if os.path.exists(self.directory):
            self.fatal(_("destination path '%s' already exists") % \
                       self.directory)

        conduit = Conduit(self.directory)
        conduit.init(self.repository,
                     bare=self.options.bare,
                     domain=self.options.domain,
                     quiet=self.options.quiet)
        try:
            conduit.fetch(limit=self.options.limit,
                          quiet=self.options.quiet,
                          verbose=self.options.verbose,
                          flush=self.options.no_skip_latest,
                          authors=self.options.authors,
                          stop_on_unknown_author=\
                              self.options.stop_on_unknown_author)

            git = conduit.git

            if not self.options.no_repack:
                git.check_command('repack', '-adF')

            head_branch = git.symbolic_ref('HEAD')
            if head_branch == 'refs/heads/master':
                if self.options.bare:
                    git.check_command('branch', '-f', 'master', conduit.branch)
                else:
                    git.check_command('reset', '-q', '--hard', conduit.branch)
        except:
            shutil.rmtree(self.directory)
            raise

        # Verify after the above rmtree, because someone likely wants
        # to inspect the repository if the verification fails.
        if self.options.verify:
            try:
                olddir = os.getcwd()
                os.chdir(git.git_work_tree)
                Verify().eval()
            finally:
                os.chdir(olddir)



"""Command to initialize a Git repository with all the required
meta-data for 'git-cvs'."""

import os.path
import sys
import time

from cvsgit.main import Command, Conduit
from cvsgit.i18n import _

class init(Command):
    __doc__ = _(
    """Initialize a Git repository to track a CVS repository.

    Usage: %prog [options] <repository> [directory]

    Initializes a Git repository and sets up the required meta-data to
    track a CVS repository.  The 'repository' argument must be a local
    file system path pointing at the actual root of a CVS repository
    or at any module directory within the repository.

    If 'directory' is omitted, the current working directory will be
    initialized instead of the one specified.
    """)

    def initialize_options(self):
        self.repository = None
        self.add_option('--bare', action='store_true', help=\
            _("Initialize a bare repository. See git-init(1)."))
        self.add_option('--domain', metavar='DOMAIN', help=\
            _("Set the 'cvs.domain' configuration option to the "
              "e-mail domain to use as a default value for unknown "
              "authors."))
        self.add_option('--quiet', action='store_true', help=\
            _("Only print error and warning messages."))

    def finalize_options(self):
        if len(self.args) < 1:
            self.usage_error(_('missing CVS repository path'))
        elif len(self.args) == 1:
            self.repository = self.args[0]
            self.directory = None
        elif len(self.args) == 2:
            self.repository = self.args[0]
            self.directory = self.args[1]
        else:
            self.usage_error(_('too many arguments'))

    def run(self):
        conduit = Conduit(self.directory)
        conduit.init(self.repository,
                     domain=self.options.domain,
                     bare=self.options.bare,
                     quiet=self.options.quiet)

if __name__ == '__main__':
    init()



"""Command to fetch unfetched revisions from the tracked CVS
repository."""

from cvsgit.main import Command, Conduit
from cvsgit.i18n import _

class fetch(Command):
    __doc__ = _(
    """Fetch unfetched revisions from CVS.

    Usage: %prog [options]

    Fetches unfetched changes from the CVS repository we are tracking,
    merges them into related changesets and imports them into Git.
    """)

    def initialize_options(self):
        self.add_option('--limit', type='int', metavar='COUNT', help=\
            _("Stop importing after COUNT new commits."))
        self.add_option('--quiet', action='store_true', help=\
            _("Only report error and warning messages."))
        self.add_option('--verbose', action='store_true', help=\
            _("Display each changeset as it is imported."))
        self.add_authors_option()
        self.add_stop_on_unknown_author_option()

    def finalize_options(self):
        if len(self.args) > 0:
            self.usage_error(_('too many arguments'))

        self.finalize_authors_option()

    def run(self):
        conduit = Conduit()
        conduit.fetch(limit=self.options.limit,
                      quiet=self.options.quiet,
                      verbose=self.options.verbose,
                      authors=self.options.authors,
                      stop_on_unknown_author=\
                          self.options.stop_on_unknown_author)

if __name__ == '__main__':
    fetch()



"""Command to pull changes from CVS into the current branch."""

from cvsgit.main import Command, Conduit
from cvsgit.i18n import _
from cvsgit.command.verify import Verify

class pull(Command):
    __doc__ = _(
    """Update the CVS tracking branch and the current HEAD

    Usage: %prog [options]

    Does what the "fetch" command does and runs "git pull" afterwards
    to update the current branch.
    """)

    def initialize_options(self):
        self.add_option('--limit', type='int', metavar='COUNT', help=\
            _("Stop importing after COUNT new commits."))
        self.add_option('--quiet', action='store_true', help=\
            _("Only report error and warning messages."))
        self.add_option('--verbose', action='store_true', help=\
            _("Display each changeset as it is imported."))
        self.add_option('--verify', action='store_true', help=\
            _("Verify the new HEAD revision and work tree against "
              "a fresh CVS checkout (does not work in a bare "
              "repository.)"))
        self.add_no_skip_latest_option()
        self.add_authors_option()
        self.add_stop_on_unknown_author_option()

    def finalize_options(self):
        if len(self.args) > 0:
            self.usage_error(_('too many arguments'))

        self.finalize_authors_option()

    def run(self):
        conduit = Conduit()
        conduit.pull(limit=self.options.limit,
                     quiet=self.options.quiet,
                     verbose=self.options.verbose,
                     flush=self.options.no_skip_latest,
                     authors=self.options.authors,
                     stop_on_unknown_author=\
                         self.options.stop_on_unknown_author)

        # Optionally verify the new HEAD revision and work tree
        # against a fresh CVS checkout.
        if self.options.verify:
            Verify().eval()

if __name__ == '__main__':
    pull()



"""Metadata for CVSGit about a pair of CVS and Git repositories."""

import os.path
import re
import sqlite3

from cvsgit.changeset import Change, ChangeSet
from cvsgit.i18n import _

class MetaDb(object):
    """Database of CVS revisions (changes) and combined changesets.

    The database contains revisions from the CVS repository that are
    pending or already grouped into changesets and a mark indicating
    whether a changeset was already processed.  For Git, the mark is
    the SHA1 commit hash.  Unprocessed changesets have the mark None.
    """

    def __init__(self, filename):
        self.filename = filename
        self._dbh = None

    def get_dbh(self):
        if self._dbh is None:
            dbh = sqlite3.connect(self.filename)

            # http://web.utk.edu/~jplyon/sqlite/SQLite_optimization_FAQ.html
            dbh.execute("PRAGMA synchronous=OFF")
            dbh.execute("PRAGMA count_changes=OFF")
            #dbh.execute("PRAGMA cache_size=4000")

            # This should also help, but causes the behaviour of the
            # ROLLBACK command to become undefined.
            dbh.execute("PRAGMA journal_mode=OFF")

            # This helps for the temporary table which we create when
            # iterating over changes for changeset generation, but
            # bloats the process image a lot.
            #dbh.execute("PRAGMA temp_store=MEMORY")

            # Create the table that contains changes pulled from CVS.
            #
            # 'changeset_id' is NULL until a change is associated with
            # a complete changeset.
            sql = 'CREATE TABLE IF NOT EXISTS change (' \
                  'timestamp DATETIME NOT NULL, ' \
                  'author VARCHAR NOT NULL, ' \
                  'log TEXT NOT NULL, ' \
                  'filestatus CHAR(1) NOT NULL, ' \
                  'filename VARCHAR NOT NULL, ' \
                  'revision VARCHAR NOT NULL, ' \
                  'state VARCHAR(8) NOT NULL, ' \
                  'mode CHAR(1) NOT NULL, ' \
                  'changeset_id INTEGER, ' \
                  'PRIMARY KEY (filename, revision))'
            dbh.execute(sql)
            sql = 'CREATE INDEX IF NOT EXISTS change__changeset_id ' \
                  'ON change (changeset_id)'
            dbh.execute(sql)
            dbh.execute("""
                CREATE INDEX IF NOT EXISTS change__timestamp
                ON change (timestamp)""")

            # Create the table that defines the attributes of complete
            # changesets.  'id' will be referenced by one or more rows
            # in the 'change' table.
            sql = 'CREATE TABLE IF NOT EXISTS changeset (' \
                  'id INTEGER PRIMARY KEY, ' \
                  'start_time DATETIME NOT NULL, ' \
                  'end_time DATETIME NOT NULL, ' \
                  'mark VARCHAR)'
            dbh.execute(sql)
            sql = 'CREATE UNIQUE INDEX IF NOT EXISTS ' \
                  'changeset__id__start_time__mark ' \
                  'ON changeset (id, start_time, mark)'
            dbh.execute(sql)

            # Create the table that stores stat() information for all
            # paths in the CVS repository.  This allows the CVS change
            # scanner to skip unmodified RCS files and directories.
            dbh.execute("""
                CREATE TABLE IF NOT EXISTS statcache (
                    path VARCHAR PRIMARY KEY,
                    mtime INTEGER NOT NULL,
                    size INTEGER NOT NULL)""")
            # With this index, I haven't observed any speed gain.
            #dbh.execute("""
            #    CREATE INDEX IF NOT EXISTS statcache_index
            #    ON statcache (path, mtime, size)""")

            self._dbh = dbh
        return self._dbh
    
    dbh = property(get_dbh)

    def load_statcache(self):
        """Load the complete stat() cache and return it as a dictionary
        of the form {path:(mtime, size)}.
        """
        sql = 'SELECT path, mtime, size FROM statcache'
        statcache = {}
        for row in self.dbh.execute(sql):
            statcache[row[0]] = row[1:]
        return statcache

    def update_statcache(self, statcache):
        """'statcache' is a dictionary of {path:(mtime, size)} to insert
        into or update in the meta database's stat() cache.
        """
        sql = 'INSERT OR REPLACE INTO statcache ' \
              '(path, mtime, size) VALUES (?,?,?)'
        for path in statcache.keys():
            values = (path,) + statcache[path]
            self.dbh.execute(sql, values)

    def add_change(self, change):
        """Insert a single file change into the database.

        If a record for the specified file and revision exists it is
        assumed to be identical and the change will be ignored.

        Note that multiple updates to the database may be grouped into
        a larger transaction for performance, but the caller can use
        commit() to ensure that the changes get flushed to disk.
        """
        self.dbh.execute("""
            INSERT OR IGNORE INTO change
                (timestamp, author, log, filestatus, filename,
                revision, state, mode)
            VALUES (?,?,?,?,?,?,?,?)""",
            (change.timestamp, change.author, change.log,
             change.filestatus, change.filename, change.revision,
             change.state, change.mode,))

    def add_changeset(self, changeset):
        """Record the attributes of 'changeset' and mark the
        referenced changes as belonging to this changeset.

        Associating changes with a changeset can be considered an
        atomic operation from the caller's perspective.
        """
        id = self.dbh.execute("""
            INSERT INTO changeset (start_time, end_time) VALUES (?,?)
            """, (changeset.start_time, changeset.end_time,)).lastrowid
        try:
            self.dbh.executemany("""
                UPDATE change SET changeset_id=%d
                WHERE filename=? AND revision=?
                """ % id, map(lambda c: (c.filename, c.revision),
                              changeset.changes))
        except:
            self.dbh.execute("""
                UPDATE change SET changeset_id=NULL
                WHERE changeset_id=%dfilename=? AND revision=?
                """ % id)
            raise

    def mark_changeset(self, id, mark):
        """Mark 'changeset' as having been integrated.
        """
        assert(id != None)
        sql = 'UPDATE changeset SET mark=? WHERE id=?'
        self.dbh.execute(sql, (mark, id,))
        self.dbh.commit()

    def begin_transaction(self):
        """Starts a new transaction (disables autocommit).
        """
        self.dbh.execute('BEGIN TRANSACTION')

    def commit(self):
        """Commits the current transaction.
        """
        self.dbh.commit()

    def end_transaction(self):
        """Ends a new transaction (enables autocommit).
        """
        self.dbh.execute('END TRANSACTION')

    def count_changes(self):
        """Return the number of free changes (not bound in a changeset).
        """
        return self.dbh.execute("""
            SELECT COUNT(*)
            FROM change
            WHERE changeset_id IS NULL""").fetchone()[0]

    def changes_by_timestamp(self, processed=None, reentrant=True):
        """Yields a list of changes recorded in the database.

        The 'processed' keyword determines wheather changes which are
        already included in a changeset are to be included or not.  If
        the value is neuter True nor False, all changes are included.
        """
        if processed == True:
            where = 'changeset_id IS NOT NULL'
        elif processed == False:
            where = 'changeset_id IS NULL'
        else:
            where = '1'

        def mkchange(row):
            return Change(timestamp=row[0], author=row[1], log=row[2],
                          filestatus=row[3], filename=row[4],
                          revision=row[5], state=row[6], mode=row[7])

        if not reentrant:
            for row in self.dbh.execute("""
                SELECT timestamp, author, log, filestatus, filename,
                       revision, state, mode
                FROM change
                WHERE %s
                ORDER BY timestamp, filename, revision""" % where):
                yield(mkchange(row))
            return

        self.dbh.execute("""
            CREATE TEMPORARY TABLE free_change AS
            SELECT timestamp, author, log, filestatus, filename,
                   revision, state, mode
            FROM change
            LIMIT 0""")
        self.dbh.execute("""
            CREATE INDEX free_change__timestamp
            ON free_change (timestamp)""")
        self.dbh.execute("""
            CREATE INDEX free_change__filename__revision
            ON free_change (filename, revision)""")
        self.dbh.execute("""
            INSERT INTO free_change
            SELECT timestamp, author, log, filestatus, filename,
                   revision, state, mode
            FROM change
            WHERE %s""" % where)

        try:
            while True:
                rows = self.dbh.execute("""
                    SELECT timestamp, author, log, filestatus, filename,
                           revision, state, mode
                    FROM free_change
                    ORDER BY timestamp
                    LIMIT 1000""").fetchall()
                if len(rows) == 0:
                    break
                for row in rows:
                    change = mkchange(row)
                    yield(change)
                    self.dbh.execute("""
                        DELETE FROM free_change
                        WHERE filename = ? AND revision = ?""",
                        (change.filename, change.revision,))
        finally:
            self.dbh.execute('DROP TABLE IF EXISTS free_change')

    def count_changesets(self):
        """Return the number of unmarked changesets (not imported).
        """
        sql = 'SELECT COUNT(*) FROM changeset WHERE mark IS NULL'
        return self.dbh.execute(sql).fetchone()[0]

    def _select_changesets(self, where):
        where = where % {'changeset':'cs'}
        sql = """
            SELECT cs.id, cs.start_time, cs.end_time, c.timestamp,
                   c.author, c.log, c.filestatus, c.filename,
                   c.revision, c.state, c.mode
            FROM changeset cs
            INNER JOIN change c ON c.changeset_id = cs.id
            WHERE %s
            ORDER BY cs.end_time, cs.id""" % where

        changeset = None
        for row in self.dbh.execute(sql):
            change = Change(timestamp=row[3],
                            author=row[4],
                            log=row[5],
                            filestatus=row[6],
                            filename=row[7],
                            revision=row[8],
                            state=row[9],
                            mode=row[10])

            if changeset is None or changeset.id != row[0]:
                if changeset:
                    yield(changeset)

                changeset = ChangeSet(change, id=row[0])
                changeset.provider = self
                changeset.start_time = row[1]
                changeset.end_time = row[2]
            else:
                changeset.changes.append(change)

        if changeset:
            yield(changeset)

    def head_changeset(self):
        """Return the "head" changeset, the one with the highest value
        of 'id' ('mark' is ignored) or None if there is no changeset.
        """
        where = '%(changeset)s.id IS MAX(%(changeset)s.id)'
        for cs in self._select_changesets(where):
            return cs

    def changesets_by_start_time(self):
        """Yield a list of all unmarked changesets currently recorded
        in the database, ordered by their start time.
        """
	where = '%(changeset)s.mark IS NULL'
        return self._select_changesets(where)

    def all_authors(self):
        """Return a list of all author login names.
        """
        return map(lambda row: row[0], self.dbh.execute("""
            SELECT DISTINCT(author) FROM change ORDER BY author
        """).fetchall())



"""Sub-command logic for git-cvs."""

import os
import string
import sys
import textwrap
from optparse import OptionParser
from cvsgit.i18n import _

class Cmd(object):
    """Abstract base class for conduit commands.
    """

    def __init__(self):
        # Parse the __doc__ string of this command class. This string
        # must contain three elements: a one-line summary, a short
        # usage help and a long description.  The order is fixed.
        docitems = self.__doc__.split('\n\n', 2)
        if len(docitems) != 3:
            raise RuntimeError, \
                'Not enough elements in __doc__ string of %s' % \
                self.__class__
        self.summary, self.usage, self.description = docitems
        self.summary = string.join(textwrap.dedent(self.summary).
                                   splitlines())
        self.usage = textwrap.dedent(self.usage)
        self.description = textwrap.dedent(self.description).rstrip()

        # For convenience, run the command immediately with arguments
        # taken from sys.argv if the command's module  is __main__.
        if self.__module__ == '__main__':
            import sys
            self.main(sys.argv)

    def _main(self, argv):
        self.option_parser = OptionParser(
            prog=os.path.basename(argv[0]),
            description=self.description,
            usage=self.usage,
            add_help_option=False)

        # Turn off GNU-style mix of arguments and options.  We have
        # subcommands that accept --help, so the main command should
        # not interpret --help after a subcommand name.
        self.option_parser.allow_interspersed_args = False

        self.add_option('--help', action='help', help=\
            _("Show this help and exit."))

        self.initialize_options()
        self.options, self.args = self.option_parser.parse_args(argv[1:])
        self.finalize_options()
        sys.exit(self.run())

    def eval(self, *args):
        try:
            prog = self.__class__.__name__
            self.main([prog] + list(args))
        except SystemExit, e:
            if e.code is None or e.code == 0:
                return 0
            else:
                return e.code

    def main(self, argv):
        try:
            self._main(argv)
        except KeyboardInterrupt:
            # In good old tradition: 128 + signal number (SIGINT=2)
            sys.exit(128 + 2)

    def add_option(self, *args, **kwargs):
        return self.option_parser.add_option(*args, **kwargs)

    def usage_error(self, msg):
        self.option_parser.error(msg)

    def warn(self, msg):
        sys.stderr.write(_('%s: warning: %s\n') % (self.option_parser.prog, msg))

    def error(self, msg):
        sys.stderr.write(_('%s: error: %s\n') % (self.option_parser.prog, msg))

    def fatal(self, msg):
        sys.stderr.write(_('%s: fatal: %s\n') % (self.option_parser.prog, msg))
        sys.exit(1)

    def initialize_options(self):
        raise RuntimeError, \
            'abstract method -- subclass %s must override' % self.__class__

    def finalize_options(self):
        raise RuntimeError, \
            'abstract method -- subclass %s must override' % self.__class__

    def run(self, *args):
        raise RuntimeError, \
            'abstract method -- subclass %s must override' % self.__class__



"""Git interface module for 'git-cvs'."""

import os
import time
import types
import shutil
import sys

from signal import signal, SIGHUP, SIGINT, SIGTERM, SIG_IGN
from subprocess import Popen, PIPE

from cvsgit.changeset import FILE_DELETED
from cvsgit.i18n import _
from cvsgit.error import Error
from cvsgit.utils import stripnl
from cvsgit.term import NoProgress

# I don't know how GIT_DIR and GIT_WORK_TREE and GIT_OBJECT_DIRECTORY
# and all the rest could affect us here, so I'll just discard them all
# for now.
def safe_environ():
    env = os.environ.copy()
    for k in env.keys():
        if k.startswith('GIT_'):
            del env[k]
    return env

class GitError(Error):
    """Base exception for errors in the cvsgit.git module"""
    pass

class GitCommandError(GitError):
    """Failure to run an external "git" command

    The 'command' attribute will be an array representing the
    command that failed and 'returncode' will contain the exit
    code of the command.  The 'stderr' member may contain the
    error output from the command, but may also be None.
    """

    def __init__(self, command, returncode, stderr=None):
        self.command = command
        self.returncode = returncode
        self.stderr = stderr
        msg = "'%s' exited with code %d" % \
            (' '.join(command), returncode)
        if stderr:
            stderr = '\n  '.join(stripnl(stderr).split('\n'))
            msg += '\n\nError output of %s command:\n  %s' % \
                (command[0], stderr)
        super(GitCommandError, self).__init__(msg)

class MissingAuthorFullname(GitError):
    """Raised when there is no known fullname for an author's login
    name and --stop-on-missing-author was given on the command-line.
    """

    def __init__(self, author):
        msg = 'missing fullname for author: %s' % author
        super(GitError, self).__init__(msg)

class Git(object):
    """Git repository and optional work tree.

    The Git repository may or may not already exist until the init()
    method is called, after which the repository will definitely
    exist, if the call returns successfully.
    """

    def __init__(self, directory=None):
        """Construct a Git repository object.
        """
        if directory == None:
            self._directory = os.getcwd()
        else:
            self._directory = directory

    def get_directory(self):
        """Return the repository top-level path.

        This may be the metadata directory or the top-level of the
        work tree depending on whether the repository is bare or not.

        This method may be called before the repository has been
        initialized and will always return the same result.
        """
        return self._directory

    directory = property(get_directory)

    def is_bare(self):
        """Return True iff the repository exists and is bare.

        This method may be called before the repository has been
        initialized, but may return a different result.
        """
        return self.config_get('core.bare') == 'true'

    def get_git_dir(self):
        """Return the path to the metadata directory.

        This method may be called before the repository has been
        initialized, but may return a different result.
        """
        if self.is_bare():
            return self.directory
        else:
            return os.path.join(self.directory, '.git')

    git_dir = property(get_git_dir)

    def get_git_work_tree(self):
        """Return the path to the working tree.

        If the repository is bare, None is returned since it has no
        associated work tree.

        This method may be called before the repository has been
        initialized, but may return a different result.
        """
        if self.is_bare():
            return None
        else:
            return self.directory

    git_work_tree = property(get_git_work_tree)

    def _popen(self, command, **kwargs):
        if not kwargs.has_key('env'):
            kwargs['env'] = safe_environ()
        if not kwargs.has_key('cwd'):
            kwargs['cwd'] = self.directory
        return Popen(command, **kwargs)

    def init(self, bare=False, quiet=False):
        """Initialize or reinitialize the repository.
        """
        args = []
        if bare:
            args.append('--bare')
        if quiet:
            args.append('--quiet')

        directory_created = False
        try:
            if not os.path.isdir(self.directory):
                os.mkdir(self.directory)
                directory_created = True

            command = ['git', 'init'] + args
            pipe = self._popen(command, stderr=PIPE)
            dummy, stderr = pipe.communicate()
            if pipe.returncode != 0:
                raise GitCommandError(command, pipe.returncode, stderr)
        except:
            if directory_created:
                shutil.rmtree(self.directory)
            raise

    def destroy(self):
        """Recursively remove the repository directory.
        """
        if os.path.exists(self.directory):
            shutil.rmtree(self.directory)

    def checkout(self, *args):
        """Call 'git checkout' with given arguments.
        """
        self.check_command('checkout', *args)

    def rebase(self, *args):
        """Call 'git rebase' with given arguments.
        """
        self.check_command('rebase', *args)

    def pull(self, *args):
        """Call 'git pull' with given arguments.
        """
        self.check_command('pull', *args)

    def rev_parse(self, *args):
        """Return the output of 'git rev-parse <*args>'
        """
        return self.check_command('rev-parse', *args, stdout=PIPE)

    def rev_list(self, *args):
        """Return the output of 'git rev-list <*args>'
        """
        return self.check_command('rev-list', *args, stdout=PIPE)

    def symbolic_ref(self, *args):
        """Return the output of 'git symbolic-ref <*args>'
        """
        return self.check_command('symbolic-ref', *args, stdout=PIPE)

    def call(self, command, *args, **kwargs):
        """Run a "git" subcommand with given arguments.

        See subprocess.call for the list of available keyword
        arguments.

        Returns the command's exit code.
        """
        command = ['git', command] + list(args)
        return subprocess.call(command, **kwargs)

    def check_command(self, command, *args, **kwargs):
        """Run a "git" subcommand with given arguments.

        Raises a GitCommandError if the subcommand does not return a
        zero exit code.

        The stdout keyword argument can take any value that
        subprocess.Popen would accept.  If it is subprocess.PIPE, then
        the output of the command is returned as a string.
        """
        stdout = None
        for kw in kwargs.keys():
            if kw == 'stdout':
                stdout = kwargs[kw]
            else:
                raise ArgumentError, 'Invalid keyword: %s' % kw

        command = ['git', command] + list(args)
        pipe = self._popen(command, stdout=stdout, stderr=PIPE)
        out, err = pipe.communicate()
        if pipe.returncode != 0:
            raise GitCommandError(command, pipe.returncode, err)
        if stdout == PIPE:
            return stripnl(out)

    def config_get(self, varname, default=None):
        """Retrieve the value of a config variable.

        This method may be called before the repository exists.  In
        that case it will always return the default value.
        """
        if not os.path.isdir(self.directory):
            return default
        command = ['git', 'config', '--get', varname]
        pipe = self._popen(command, stdout=PIPE, stderr=PIPE)
        stdout, stderr = pipe.communicate()
        if pipe.returncode == 0:
            return stripnl(stdout)
        elif pipe.returncode == 1:
            return default
        else:
            raise GitCommandError(command, pipe.returncode, stderr)

    def config_set(self, varname, value):
        """Set the value of a config variable.
        """
        self.check_command('config', varname, value)

    def import_changesets(self, changesets, branch, domain=None,
                          limit=None, verbose=False,
                          progress=None, total=None,
                          authors=None, stop_on_unknown_author=False):
        """Loop over changesets and import them.
        """
        if progress == None:
            progress = NoProgress()
        with progress:
            self._import_changesets(changesets, branch, domain,
                                    limit, verbose, progress,
                                    total, authors, stop_on_unknown_author)

    def _import_changesets(self, changesets, branch, domain, limit,
                           verbose, progress, total, authors,
                           stop_on_unknown_author):
        message = _('Importing changesets')
        def do_progress(count, total):
            progress(message, count, total)

        class SignalIndicator():
            def __init__(self):
                self.signal = {}
            def __call__(self, signalnum, frame):
                self.signal[signalnum] = True
            def isset(self, signalnum=None):
                if signalnum:
                    return self.signal.has_key(signalnum)
                else:
                    return len(self.signal) > 0

        sigaction = SignalIndicator()
        signalset = (SIGHUP,SIGINT,SIGTERM,)
        old_sigaction = {}
        for signalnum in signalset:
            old_sigaction[signalnum] = signal(signalnum, sigaction)

        def ignore_signals():
            for signalnum in signalset:
                signal(signalnum, SIG_IGN)

        # absolulte file name since cwd is changed in _popen
        marksfile = os.path.join(self.git_dir, 'cvsgit.marks')
        marksfile = os.path.abspath(marksfile)

        command = ['git', 'fast-import', '--quiet']
        command.append('--export-marks=' + marksfile)
        pipe = self._popen(command, stdin=PIPE, preexec_fn=ignore_signals)

        if limit != None and total != None and total > limit:
            total = limit

        fi = GitFastImport(pipe, branch, domain=domain, verbose=verbose,
                           authors=authors, stop_on_unknown_author=\
                               stop_on_unknown_author)
        changeset_ids = []
        db = None
        try:
            for changeset in changesets:
                if limit != None and len(changeset_ids) >= limit:
                    break

                fi.add_changeset(changeset)
                if db == None: db = changeset.provider.metadb # FIXME
                changeset_ids.append(changeset.id)
                do_progress(len(changeset_ids), total)

                if sigaction.isset(SIGINT):
                    raise KeyboardInterrupt()
                elif sigaction.isset():
                    break
        finally:
            try:
                fi.close()
            finally:
                self.mark_changesets(db, changeset_ids)
                for signalnum in signalset:
                    signal(signalnum, old_sigaction[signalnum])

        if fi.returncode != 0:
            raise RuntimeError, _('git fast-import failed')

    def mark_changesets(self, db, changeset_ids):
        filename = os.path.join(self.git_dir, 'cvsgit.marks')
        if not os.path.isfile(filename):
            return

        marks = {}
        f = file(filename, 'r')
        try:
            for line in f.readlines():
                mark, sha1 = line.rstrip().split()
                marks[int(mark[1:])] = sha1
        finally:
            f.close()

        for id in changeset_ids:
            if marks.has_key(id):
                sha1 = marks[id]
                db.mark_changeset(id, sha1)

class GitFastImport(object):
    def __init__(self, pipe, branch, domain=None, verbose=False,
                 authors=None, stop_on_unknown_author=False):
        self.pipe = pipe
        self.branch = branch
        self.domain = domain
        self.verbose = verbose
        self.authors = authors
        self.stop_on_unknown_author = stop_on_unknown_author
        self.last_changeset = None

    def add_changeset(self, changeset):
        name = self.author_name(changeset.author)
        email = self.author_email(changeset.author)
        when = self.raw_date(changeset.timestamp)

        if self.verbose:
            tstamp = time.strftime('%Y-%m-%d %H:%M:%S',
                time.gmtime(changeset.timestamp))
            print '[%d] %s %s' % (changeset.id, tstamp, name)
            teaser = changeset.log.splitlines()[0]
            if len(teaser) > 68:
                teaser = teaser[:68] + '...'
            print '\t%s' % teaser.encode('ascii', 'replace')

        self.write('commit %s\n' % self.branch)
        self.write('mark :%s\n' % changeset.id)
        self.write('committer %s <%s> %s\n' % (name, email, when))
        self.data(changeset.log.encode('utf-8'))

        if changeset.id != 1:
            # FIXME: this is a hack; find out if the branch exists
            if self.last_changeset is None:
                self.write('from %s^0\n' % self.branch)
            else:
                self.write('from :%s\n' % self.last_changeset.id)
        self.last_changeset = changeset

        for c in changeset.changes:
            if self.verbose:
                print '\t%s %s %s' % (c.filestatus, c.filename, c.revision)
            if c.filestatus == FILE_DELETED:
                self.write('D %s\n' % c.filename)
            else:
                blob = changeset.blob(c)
                perm = changeset.perm(c)

                # Git according to git-fast-import(1) only supports
                # these two file modes for plain files.
                if (perm & 0111) != 0:
                    perm = 0755
                else:
                    perm = 0644

                self.write('M %o inline %s\n' % (perm, c.filename))
                self.data(blob)

    def close(self):
        try:
            self.pipe.stdin.close()
            self.pipe.wait()
        except:
            pass
        self.returncode = self.pipe.returncode

    def raw_date(self, seconds):
        """Convert 'seconds' from seconds since the epoch to Git's
        native date format."""
        return '%s +0000' % (seconds)

    def author_name(self, author):
        # TODO: allow author name mappings to be defined for a
        # certain time span, since CVS committer name could be
        # reused.
        if self.authors:
            if self.authors.has_key(author):
                return self.authors[author][0]
            elif self.stop_on_unknown_author:
                raise MissingAuthorFullname(author)
        return author

    def author_email(self, author):
        if (self.authors and
                self.authors.has_key(author) and
                self.authors[author][1] is not None):
            return self.authors[author][1]
        elif self.domain:
            return '%s@%s' % (author, self.domain)
        else:
            return author

    def data(self, data):
        "'data' must be a raw binary string of the str() type."
        assert type(data) == types.StringType, \
            "data type is %s" % type(data)
        self.write('data %d\n' % len(data))
        self.write(data)
        self.write('\n')

    def write(self, data):
        self.pipe.stdin.write(data)



"""Command to draw the revision tree of an RCS file."""

import os

from cvsgit.main import Command
from cvsgit.rcs import RCSFile
from cvsgit.i18n import _

# FIXME: RCS should expand keywords, not CVS
from cvsgit.cvs import CVS

class Rcsdump(Command):
    __doc__ = _(
    """Dump all changes on the HEAD branch of an RCS file.

    Usage: %prog [options] <rcsfile>

    Displays the changes on the HEAD branch of an RCS file for
    debugging purposes.
    """)

    def initialize_options(self):
        self.add_option('--checkout', metavar='REVISION', help=\
            _("Dump the content of the specified REVISION."))

    def finalize_options(self):
        if len(self.args) < 1:
            self.usage_error(_('missing RCS file argument'))
        elif len(self.args) == 1:
            self.rcsfile = self.args[0]
        else:
            self.usage_error(_('too many arguments'))

    def run(self):
        rcsfile = RCSFile(self.rcsfile)

        if self.options.checkout:
            return self.checkout(rcsfile, self.options.checkout)

        print 'Head: %s' % rcsfile.head
        print 'Branch: %s' % rcsfile.branch
        print 'Revision trail:',
        for revision in rcsfile.revisions():
            print revision,
        print ''
        for change in rcsfile.changes():
            rcsfile._print_revision(change.revision)

    def checkout(self, rcsfile, revision):
        change = rcsfile.change(revision)
        blob = rcsfile.blob(revision)
        cvs = CVS(os.path.join(os.path.dirname(rcsfile.filename)), None)
        print cvs.expand_keywords(blob, change, rcsfile, revision)



"""Command to fetch changes from the CVS repository."""

import re
import subprocess
from subprocess import PIPE
import sys

from cvsgit.cvs import split_cvs_source
from cvsgit.i18n import _
from cvsgit.main import Command, Conduit
from cvsgit.utils import Tempdir, stripnl
from cvsgit.term import NoProgress, Progress

class FetchChanges(Command):
    __doc__ = _(
    """Fetch changes from CVS

    Usage: %prog

    Fetches unseen changes from the CVS repository and stores them
    locally for changeset generation at a later point in time.
    """)

    def initialize_options(self):
        self.add_quiet_option()

    def finalize_options(self):
        if len(self.args) > 0:
            self.usage_error(_('too many arguments'))

        if self.options.quiet:
            self.progress = NoProgress()
        else:
            self.progress = Progress()

    def run(self):
        conduit = Conduit()
        cvs = conduit.cvs
        cvs.fetch_changes(progress=self.progress)



"""Changeset reconstruction logic for CVSGit."""

QUIET_PERIOD = 60

FILE_ADDED = 'A'
FILE_DELETED = 'D'
FILE_MODIFIED = 'M'

class Change(object):
    """Representation of a single change in an RCS file.

    For example:
    >>> Change(1303768248, "jack", "ToDo list added", FILE_ADDED,
    ... "todo.txt", "1.1", "state???", "mode???").author
    'jack'

    The time stamp should be in UTC time zone. The state and mode
    arguments in the above example are bogus, which illustrates
    that this class is really just a dumb container.

    Change objects are integrated into a ChangeSet by a
    ChangeSetGenerator."""

    def __init__(self, timestamp, author, log, filestatus, filename,
                 revision, state, mode):
        self.timestamp = timestamp
        self.author = author
        self.log = log
        self.filestatus = filestatus
        self.filename = filename
        self.revision = revision
        self.state = state
        self.mode = mode

    def __str__(self):
        return '<%s %s, %s, %s %s %s %s>' % \
            (type(self).__name__, self.timestamp, self.author,
             self.filestatus, self.filename, self.revision, self.state)

class ChangeSet(object):
    """A set of Change objects that represent a CVS commit.

    The ChangeSet class represents a set of related RCS changes
    that appear to have been made in the same CVS commit.  The
    ChangeSetGenerator class constructs instances of ChangeSet.

    One or more Change objects can be integrated into a ChangeSet
    like so:

    >>> c1 = Change(1303768245, "jack", "Initial commit", FILE_ADDED,
    ... "todo.txt", "1.1", "state???", "mode???")
    >>> c2 = Change(1303768249, "jack", "Initial commit", FILE_ADDED,
    ... "README", "1.1", "state???", "mode???")
    >>> cs = ChangeSet(c1)
    >>> cs.integrate(c2)
    True
    >>> cs.log
    'Initial commit'
    >>> cs.start_time < cs.end_time
    True

    The following is a little odd, but currently wanted behaviour in
    order to make it easier to emulate a peculiarity of cvs.  The
    timestamp is always one second ahead of end_time:

    >>> cs.timestamp > cs.end_time
    True
    >>> cs.timestamp
    1303768250

    The timestamp returned for the changeset is based on the timestamp
    of the last change in the set because that is most likely the time
    when the commit actually completed and "cvs -D <timestamp>" can be
    used to create an equivalent working copy from CVS.
    """

    def __init__(self, change, id=None, mark=None, provider=None):
        """'id' is an arbitrary value to distinguish this changeset.

        'mark' is a commit marker that, if not None, identifies this
        changeset as having been integrated.

        The 'provider' argument, if set, should be an object that has
        a blob() method accepting a Change object as an argument and
        returning the binary data for the file as it was after the
        change."""

        self.id = id
        self._mark = mark
        self._provider = provider
        self.start_time = change.timestamp
        self.end_time = change.timestamp
        self.changes = [change]

    def get_provider(self):
        if self._provider is None:
            raise RuntimeError, \
                'no provider has been set for this changeset'
        return self._provider

    def set_provider(self, provider):
        self._provider = provider

    provider = property(get_provider, set_provider, None,
                        'source VCS of this changeset')

    def get_mark(self):
        return self._mark

    def set_mark(self, mark):
        self._mark = mark
        self.provider.mark_changeset(self)

    mark = property(get_mark, set_mark, None,
                    """anything but None marks the changeset as
                    integrated into the target VCS""")

    def perm(self, change):
        return self.provider.perm(change)

    def blob(self, change):
        return self.provider.blob(change, self)

    def get_timestamp(self):
        # At first, this method returned start_time, but it makes more
        # sense to return end_time, which is when the last RCS change
        # happend that affected this changeset.  This should make it
        # more likely that a "cvs checkout -D <timestamp>" yields the
        # expected result when <timestamp> is that of the changeset.

        # FIXME: end_time + 1 is a lie
        # It is however an easy workaround to make the initial
        # revisions be 1.1 instead of 1.1.1.1 (consistently, in both
        # "cvs co -D <date>" and in the cloned Git repository.)
        return self.end_time + 1

    def get_author(self):
        return self.changes[0].author

    def get_log(self):
        return self.changes[0].log

    def get_filenames(self):
        return map(lambda c: c.filename, self.changes)

    timestamp = property(get_timestamp)
    author = property(get_author)
    log = property(get_log)
    filenames = property(get_filenames)

    def integrate(self, change):
        if change.author != self.author or \
           change.log != self.log or \
           change.filename in self.filenames:
            return False

        if change.timestamp < self.start_time:
            self.start_time = change.timestamp
        elif change.timestamp > self.end_time:
            self.end_time = change.timestamp

        self.changes.append(change)
        return True

    def __str__(self):
        if len(self.changes) == 1:
            changes = '1 file'
        else:
            changes = '%d files' % len(self.changes)
        return '<%s %s, %s, %s, %s>' % \
            (type(self).__name__, self.timestamp, self.author, changes,
             self.mark)

class ChangeSetGenerator(object):
    """Group a series of individual file changes into changesets that
    have likely been committed together.  The individual changes must
    be presented in ascending order of their timestamp."""

    def __init__(self, quiet_period=QUIET_PERIOD, limit=None):
        """Construct a new ChangeSetGenerator instance.

        The ChangeSetGenerator should only return complete ChangeSets
        and retain possibly incomplete ChangeSets.  If limit is given,
        only return as many changesets and retain all others.

        Major reasons for incomplete ChangeSets:

        1. When importing Changes directly from a writable CVS
           repository (without locking against CVS clients) we could
           be racing against "cvs commit" and other operations.

        2. When importing from a CVS repository mirror, the mirroring
           may have overlapped with a "cvs commit" or other operation,
           again unless CVS clients were locked out for that time.

        3. If the CVS mirror operation was interrupted, some RCS files
           belonging to a ChangeSet may have been updated while others
           have not.

        To guard against all of the above, the ChangeSetGenerator
        defines a ChangeSet `X' as being complete only if there is
        another ChangeSet `Y' and `Y.start_time - X.end_time >
        quiet_period`, i.e., after the complete ChangeSet there is at
        least one significantly younger one.
        """

        self.quiet_period = quiet_period
        self.limit = limit
        self.count = 0
        self.changesets = []

    def integrate(self, change):
        """Integrate a single file change into the an appropriate
        changeset and yield changesets that haven't been modified for
        at least the "quiet period", relative to the given change."""

        # Yield changesets that have passed the "quiet period".
        changesets = []
        count = self.count
        for cs in self.changesets:
            delta = change.timestamp - cs.end_time
            if delta >= self.quiet_period:
                if self.limit and count >= self.limit:
                    return
                count += 1
                yield(cs)
            else:
                changesets.append(cs)
        self.changesets = changesets
        self.count = count

        # For all remaining changesets, try to find one that can
        # integrate the change.  Otherwise, open a new changeset.
        for cs in self.changesets:
            if cs.integrate(change):
                return
        self.changesets.append(ChangeSet(change))

        # TODO: Is changeset ordering more stable with this?
        #self.changesets.sort(key=lambda cs: cs.timestamp)

    def flush(self):
        """Yield remaining changesets up to the limit (if one was set),
        even potentially incomplete ones."""

        count = self.count
        for cs in self.changesets:
            if self.limit and count >= self.limit:
                return
            count += 1
            yield(cs)
        self.changesets = []
        self.count = count



"""Utility functions and classes."""

import os
import shutil
import tempfile
import textwrap

class Tempdir(object):
    """Manage a tempoarary directory

    >>> with Tempdir() as tempdir:
    ...   os.mkdir(os.path.join(tempdir, 'blabla'))
    """

    def __init__(self, cwd=False):
        """Initialize a Tempdir context

        If 'cwd' is True, make the temporary directory the current
        working directory when the context is entered and restore
        the previous working directory when the context is exited.
        """
        self.cwd = cwd

    def __enter__(self):
        self.tempdir = tempfile.mkdtemp()
        if self.cwd:
            self.pwd = os.getcwd()
            os.chdir(self.tempdir)
        return self.tempdir

    def __exit__(self, exception_type, value, traceback):
        if self.cwd:
            os.chdir(self.pwd)
        if os.path.isdir(self.tempdir):
            shutil.rmtree(self.tempdir)
        return False

class Environ(object):
    """Environment variable context

    >>> with Environ(PATH='/bin:/usr/bin'):
    ...   pass
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        self.oldenv = os.environ.copy()
        for k in self.kwargs.keys():
            v = self.kwargs[k]
            if v == None:
                if os.environ.has_key(k):
                    del os.environ[k]
            else:
                os.environ[k] = v
        return os.environ

    def __exit__(self, exception_type, value, traceback):
        os.environ = self.oldenv
        return False

def stripnl(string):
    """Strip the final newline from <string>.

    It is an error if <string> does not end with a newline character,
    except for the empty string, which is okay.

    >>> stripnl('')
    ''
    >>> stripnl('he\\nllo\\n')
    'he\\nllo'
    >>> try:
    ...   stripnl('hello')
    ... except RuntimeError as e:
    ...   e.message
    "string doesn't end in newline: hello"
    """
    if string == '':
        return string
    elif string.endswith('\n'):
        return string[0:-1]
    else:
        raise RuntimeError("string doesn't end in newline: %s" % string)

def dedent(docstring):
    """Dedent a Python docstring treating the first line special.

    This function uses textwrap.dedent() but assumes that the first
    line is not indented.
    """
    lines = docstring.splitlines()
    return '\n'.join((lines[0], textwrap.dedent('\n'.join(lines[1:])),))



"""Command to compare a Git tree against CVS."""

import os
import re
import subprocess
from subprocess import PIPE
import sys

from cvsgit.cvs import split_cvs_source
from cvsgit.git import GitCommandError
from cvsgit.i18n import _
from cvsgit.main import Command, Conduit
from cvsgit.utils import Tempdir, stripnl

class Verify(Command):
    __doc__ = _(
    """Verify the Git work tree against CVS.

    Usage: %prog

    Compares the work tree against a clean checkout from CVS with the
    same timestamp as the HEAD commit.
    """)

    def initialize_options(self):
        self.commit = 'HEAD'
        self.add_option('--history', action='store_true', help=\
            _("Walk backwards through the entire history."))
        self.add_option('--forward', action='store_true', help=\
            _("Modifies the --history option to walk forward."))
        self.add_option('--skip', action='store_true', help=\
            _("Skip the first verification step and move HEAD "
              "backward instead, the opposite direction with "
              "--forward)."))
        self.add_option('--quiet', action='store_true', help=\
            _("Only report error and warning messages."))

    def finalize_options(self):
        if len(self.args) > 0:
            self.usage_error(_('too many arguments'))

    def run(self):
        conduit = Conduit()
        self.cvsroot, self.module = split_cvs_source(conduit.source)
        self.git = git = conduit.git
        with Tempdir() as tempdir:
            self.tempdir = tempdir

            while True:
                if self.options.skip:
                    self.options.skip = False
                else:
                    returncode = self._run()
                    if returncode != 0:
                        return returncode
                    elif not self.options.history:
                        return 0

                if self.options.forward:
                    wc_head = self.git.rev_parse('HEAD')
                    cvs_head = self.git.rev_parse(conduit.branch)
                    if wc_head == cvs_head:
                        return 0
                    subprocess.check_call(['/bin/sh', '-c',
                       'git checkout `git rev-list %s..%s|tail -1`' % \
                       (wc_head, cvs_head)])
                else:
                    try:
                        self.git.checkout('-q', 'HEAD~1')
                    except GitCommandError:
                        return 0

    def _run(self):
        returncode = 0
        date = self.commit_date(self.commit)
        head = self.git.rev_parse('--short', 'HEAD')
        command = ['cvs', '-Q', '-d', self.cvsroot, 'checkout',
                   '-P', '-D', date, '-d', self.tempdir, self.module]
        if not self.options.quiet:
            print "(%s) '%s'" % (head, "' '".join(command))
        # At least with OpenBSD's version of GNU CVS it is not
        # possible to run "cvs checkout" from a directory that's
        # inside the CVS repository.  This should avoid the
        # issue... *grrrr* The error is: cvs [checkout aborted]:
        # Cannot check out files into the repository itself
        try:
            owd = os.getcwd()
            os.chdir('/')
            subprocess.check_call(command)
        finally:
            os.chdir(owd)
        command = ['diff', '-r', self.tempdir, self.git.git_work_tree]
        pipe = subprocess.Popen(command, stdout=PIPE, stderr=PIPE)
        stdout, dummy = pipe.communicate()
        for line in stripnl(stdout).split('\n'):
            if re.match('^Only in .+: \.git$', line):
                continue
            if re.match('^Only in .+: CVS$', line):
                continue
            if not self.options.quiet:
                sys.stdout.write(line + '\n')
            returncode = 1
        return returncode

    def commit_date(self, commit):
        """Get the commit date as a string in UTC timezone.
        """
        command = ['git', 'log', '-1', commit]
        pipe = self.git._popen(command, stdout=PIPE, stderr=PIPE)
        stdout, stderr = pipe.communicate()
        if pipe.returncode != 0:
            self.fatal(_("can't get commit date of %s" % commit))
        match = re.search('Date: +[^ ]+ (.*) \+0000', stdout)
        if match:
            return '%s UTC' % match.group(1)
        else:
            self.fatal(_("couldn't match Date: in output of '%s'") % \
                           ' '.join(command))
